package ISY::Simple;;
#require 5.000;
use 5.006;

use strict;
#use warnings;

require Exporter;


# debug:
# 0x01 = report loads
# 0x02 = report urls used
# 0x04 = print func calls
# 0x08 = Dump loaded data
# 0x10 = Dump loaded data
# 0x20 =
# 0x40 =
# 0x80 =
#


our @ISA = qw(Exporter);
our %EXPORT_TAGS = ( 'all' => [ qw() ] );
our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );
our @EXPORT = qw( );

our $VERSION = '0.02';

use LWP::UserAgent;
use XML::Simple;

use Data::Dumper;
$Data::Dumper::Indent = 1;         # pretty print with array indices
$Data::Dumper::Sortkeys  = 1;         # sort indices


my %featureids = (
    'BROADBAND_SEP DEVICE' => '21080',
    'CURRENT_COST_METER' => '21014',
    'ELECTRICITY_METER' => '21011',
    'GAS_METER' => '21012',
    'NETWORKING_MODULE' => '21040',
    'NETWORK_MODULES' => '21030',
    'OPEN_AUTO-DR' => '21010',
    'PORTAL_INTEGRATION-CHECK-IT.CA' => '21070',
    'PORTAL_INTEGRATION-GREENNET.COM' => '21071',
    'SMARTMETER_ESP' => '21051',
    'UTILITY_METER_ELECTRICITY' => '21050',
    'WATER_METER' => '21013',
    'WEATHER_INFORMATION' => '21020',
    'X10' => '21060',
);


sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self = bless {}, $class;
    my %args = @_;

    my $host = $args{host};
    if (!$host && exists $ENV{"ISY_ADDR"}) {
	$host = $ENV{"ISY_ADDR"};
    }

    # gross, one of the many things that is simpler in C
    # parse   ISY_AUTH as   LOGIN:PASS
    # Over ride default login auth with ENV
    my $u = "admin";
    my $p = "admin";
    if ( exists $ENV{ISY_AUTH} ) {
	my $i = index($ENV{ISY_AUTH}, ":");
	if ( $i == -1 ) {
	    $p =  $ENV{ISY_AUTH};
	} elsif ( $i == 0 ) {
	    $p = substr($ENV{ISY_AUTH}, ++$i );
	} else {
	    $u = substr($ENV{ISY_AUTH}, 0, $i);
	    ++$i;
	    if ( $i < length($ENV{ISY_AUTH}) ) {
		$p = substr($ENV{ISY_AUTH}, $i );
	    }
	}
    }
    $self->{user} = $args{user} || $u;
    $self->{pass} = $args{pass} || $p;

    $self->{baseurl} = "http://" . $args{host} . "/rest";
    $self->{weburl} = "http://" . $args{host} . "/WEB";
    # $self->{unique_names} = $args{unique_names};
    $self->{error} = undef;
    $self->{verbose} = $args{verbose} || 0;
    $self->{debug} = $args{debug} || 0;
    $self->{cachetimeout} = 30;

    # if ( -d $HOME/.isy ) 

    $self->{ua} = LWP::UserAgent->new(
	agent => "Mozilla/5.0",
	timeout => 200,
    );

    $self->{ua}->credentials( $args{host} . ":80", '/', $self->{user}, $self->{pass});

    unless ( $args{faststart}) {
	$self->load_conf();
	$self->load_nodes();
	$self->load_prog();
	# $self->load_netw();
	# $self->load_clim();
	# $self->load_vars();
    }

    return $self;
}

sub DESTROY { }


sub flush_cache {
    my $self = shift;
    my $timeout = shift;
    my $flushall = 0;

    $timeout = $self->{cachetimeout} unless defined $timeout;

    $flushall=1 if ($timeout == 0 );

    my $curtime = time;

    if ( $flushall || ($self->{cachetime}->{programs} + $timeout) < $curtime ) {
	undef $self->{programs};
	undef $self->{prog_names};
    }

    if ( $flushall || ($self->{cachetime}->{networking} + $timeout) < $curtime ) {
	undef $self->{networking};
    }

    if ( $flushall || ($self->{cachetime}->{climate} + $timeout) < $curtime ) {
	undef $self->{climate};
    }
#  This does not expire
#
#    if ( $flushall || ($self->{cachetime}->{controls} + $timeout) < $curtime ) {
#	undef $self->{controls};
#	undef $self->{commands};
#    }

    if ( $flushall || ($self->{cachetime}->{nodes} + $timeout) < $curtime ) {
	undef $self->{nodes};
	undef $self->{node_names};
	undef $self->{node_x10};
	undef $self->{node_x10_names};
	undef $self->{scenes};
	undef $self->{scene_names};
    }

    if ( $flushall || ($self->{cachetime}->{vars} + $timeout) < $curtime ) {
	undef $self->{vars};
	undef $self->{vars_names};
    }

    return $self;
}

# for debuging
sub _dumpself {
    my $self = shift;	

    print "---dumpself--- \n";

    $self->load_conf() unless exists  $self->{commands};
    $self->load_nodes() unless exists  $self->{node_names};
    $self->load_prog() unless exists  $self->{prog_names};
    $self->load_netw() unless exists  $self->{networking};
    $self->load_clim() unless exists  $self->{climate};
    $self->load_vars();
    $self->load_log_ids();
    $self->load_node_types();

    print "=== self ===\n", Dumper($self), "\n";
}



sub load_prog {
    my $self = shift;	
    my $req;
    my $res;

    $self->{error} = undef;

    print "---load_prog--- \n" if $self->{debug} & 0x01;

    my $xurl = $self->{baseurl} . "/programs?subfolders=true";
    print "xurl = $xurl\n", if $self->{debug} & 0x02;

    $req = HTTP::Request->new(GET => $xurl);
    $req->header('Accept' => 'text/xml,application/xhtml+xml,application/xml;q=0.9,*/*;q=0');
    # $req->authorization_basic($self->{user}, $self->{pass});
    $res = $self->{ua}->request($req);
    if (! $res->is_success) {
	$self->{error} = "FAIL : load_prog : " . $res->code . " : " . $res->status_line;
	print $self->{error}, "\n";
	return undef;
    }
    my $ref = XMLin( ${$res->content_ref},  keyattr => { program => "+id" }  );

    $self->{programs} = $ref->{program};

    delete $self->{prog_names} if exists $self->{prog_names};  # fresh slate if we are reloading

    for my $key (keys (  %{$self->{ programs }} ) ) {
	my $k = uc($self->{programs}->{$key}->{name});
	$self->{prog_names}->{$k} = $key;
	#print "$k = $key\n";
    }

    $self->{cachetime}->{programs} = time;

    # print "=== Programs ===\n", Dumper($self->{programs}), "\n";

    return $self;
}

sub load_netw {
    my $self = shift;	
    my $req;
    my $res;

#    unless ( $self->check_feature('NETWORKING_MODULE') ) {
#	$self->{error} = "Networking Module module not available";
#	return  undef;
#    }

    print "---load_netw---\n" if $self->{debug} & 0x01;

    for my $s ( "wol", "resources" ) {

	my $xurl = $self->{baseurl} . "/networking/$s";
	print "xurl = $xurl\n", if $self->{debug} & 0x02;

	$req = HTTP::Request->new(GET => $xurl);
	$req->header('Accept' => 'text/xml,application/xhtml+xml,application/xml;q=0.9,*/*;q=0');
	$res = $self->{ua}->request($req);
	if (! $res->is_success) {
	    $self->{error} = "FAIL : networking/$s : " . $res->code . " : " . $res->status_line;
	    return undef;
	}
	if ( ! length( ${$res->content_ref} ) ) {
	    $self->{error} = "FAIL : networking/$s : No data";
	    return undef;
	}
	my $ref = XMLin( ${$res->content_ref},  keyattr => { NetRule => "+id" }, ForceArray => [ 'NetRule' ] );

	# print "=== networking/$s ===\n", Dumper($ref), "\n", ${$res->content_ref}, "\n";

	if ( exists $ref->{NetRule} ) {
	    $self->{networking}->{$s} = $ref->{NetRule};
	    my $n = $s . "_names";

	    for my $k ( keys %{$ref->{NetRule}} ) {
		$self->{networking}->{$n}->{ $ref->{NetRule}->{$k}->{name} } = $k;
	    }
	} else {
	    $self->{networking}->{$s} = {};
	    $self->{networking}->{$s .  "_names"} = {};
	}
    }

    $self->{cachetime}->{networking} = time;

    # print "=== networking ===\n", Dumper($self->{networking}), "\n";

    return $self;
} # netw


sub load_clim {
    my $self = shift;	
    my $req;
    my $res;

    # unless ( $self->check_feature('WEATHER_INFORMATION') ) {
# 	$self->{error} = "Weather Information module not available";
# 	return  undef;
 #    }

    print "---load_clim---\n" if $self->{debug} & 0x01;

    $self->{error} = undef;

    my $xurl = $self->{baseurl} . "/climate";
    print "xurl = $xurl\n", if $self->{debug} & 0x02;

    $req = HTTP::Request->new(GET => $xurl);
    $req->header('Accept' => 'text/xml,application/xhtml+xml,application/xml;q=0.9,*/*;q=0');
    $res = $self->{ua}->request($req);
    if (! $res->is_success) {
	$self->{error} = "FAIL : climate : " . $res->code . " : " . $res->status_line;
	print $self->{error}, "\n" if $self->{verbose};
	return undef;
    }
    if ( ! length( ${$res->content_ref} ) ) {
	$self->{error} = "FAIL : climate : No data";
	return undef;
    }
    my $ref = XMLin( ${$res->content_ref});


     # print "=== Climate ===\n", Dumper($ref), "\n";

    $self->{climate} = $ref;
    $self->{cachetime}->{climate} = time;

    return $self;
} # clim


sub load_log_ids {
    my $self = shift;	
    my $req;
    my $res;

    $self->{error} = undef;

    print "---load_log_ids---\n" if $self->{debug} & 0x01;

    my $xurl = $self->{weburl} . "/loginfo.xml";
    print "xurl = $xurl\n", if $self->{debug} & 0x02;

    $req = HTTP::Request->new(GET => $xurl);
    $req->header('Accept' => 'text/xml,application/xhtml+xml,application/xml;q=0.9,*/*;q=0');
    $res = $self->{ua}->request($req);
    if (! $res->is_success) {
	$self->{error} =  "FAIL : WEB/loginfo.xml : " . $res->code . " : " . $res->status_line;
	print $self->{error}, "\n" if $self->{verbose};
	return undef;
    }

    my $keyattr = { 
		# 'actors' => "id",
		'actor' => "id",
		'msg' => "id",
		'type' => "id",
		# 'logEntryTypes' => "id",
		# 'errorMessages' => "id",
	    };

    my $ref = XMLin( ${$res->content_ref}, keyattr => $keyattr, ForceArray => [ 'actor' ] );

    print "=== WEB/loginfo.xml ===\n", Dumper($ref), "\n" if $self->{debug} & 0x08;

    $self->{logactors} = $ref->{actors}->{actor};
    $self->{errorMsg} = $ref->{errorMessages}->{msg};
    $self->{logtype} = $ref->{logEntryTypes}->{type};


}

sub load_node_types {
    my $self = shift;	
    my $req;
    my $res;

    $self->{error} = undef;

    print "---load_node_types---\n" if $self->{debug} & 0x01;

    my $xurl = $self->{weburl} . "/cat.xml";
    print "xurl = $xurl\n", if $self->{debug} & 0x02;

    $req = HTTP::Request->new(GET => $xurl);
    $req->header('Accept' => 'text/xml,application/xhtml+xml,application/xml;q=0.9,*/*;q=0');
    $res = $self->{ua}->request($req);
    if (! $res->is_success) {
	$self->{error} =  "FAIL : WEB/cat.xml : " . $res->code . " : " . $res->status_line;
	print $self->{error}, "\n" if $self->{verbose};
	return undef;
    }

    my $keyattr = { 
		# 'NodeCategories' => "id",
		'nodeCategory' => "id",
		'nodeSubCategory' => "id",
	    };

    my $ref = XMLin( ${$res->content_ref}, keyattr => $keyattr );
    # ValueAttr => ['name']
    # ForceArray => [ 'property', 'folder' ]);

    print "=== WEB/cat.xml ===\n", Dumper($ref), "\n" if $self->{debug} & 0x08;

    $self->{nodeCategory} = $ref->{nodeCategory};

    $xurl = $self->{weburl} . "/1_fam.xml";
    print "xurl = $xurl\n", if $self->{debug} & 0x02;

    $req = HTTP::Request->new(GET => $xurl);
    $req->header('Accept' => 'text/xml,application/xhtml+xml,application/xml;q=0.9,*/*;q=0');
    $res = $self->{ua}->request($req);
    if (! $res->is_success) {
	$self->{error} =  "FAIL : WEB/1_fam.xml : " . $res->code . " : " . $res->status_line;
	print $self->{error}, "\n" if $self->{verbose};
	return undef;
    }

    $ref = XMLin( ${$res->content_ref}, keyattr => $keyattr );

    print "=== WEB/1_fam.xml ===\n", Dumper($ref), "\n" if $self->{debug} & 0x08;

    $self->{nodeFam} = $ref->{nodeCategory};

}
sub load_conf {
    my $self = shift;	
    my $req;
    my $res;

    $self->{error} = undef;

    print "---load_conf---\n" if $self->{debug} & 0x01;

    my $xurl = $self->{baseurl} . "/config";
    print "xurl = $xurl\n", if $self->{debug} & 0x02;

    $req = HTTP::Request->new(GET => $xurl);
    $req->header('Accept' => 'text/xml,application/xhtml+xml,application/xml;q=0.9,*/*;q=0');
    $res = $self->{ua}->request($req);
    if (! $res->is_success) {
	$self->{error} =  "FAIL : Controls : " . $res->code . " : " . $res->status_line;
	print $self->{error}, "\n" if $self->{verbose};
	return undef;
    }
    my $ref = XMLin( ${$res->content_ref},  keyattr => { control => "+label", feature => "+desc"}  );

    print "=== config/controls ===\n", Dumper($ref), "\n" if $self->{debug} & 0x08;


    my $cref = $ref->{controls}->{control};

    for my $key (keys (  %{$ref->{controls}->{control}} ) ) {
	my $k = uc($key);
	#print "$k\n";
	# print "-- $ref->{controls}->{control}->{$key}->{name}\n" if $self->{debug};
	$self->{controls}->{$k} = $ref->{controls}->{control}->{$key}->{name};
	$self->{commands}->{  $ref->{controls}->{control}->{$key}->{name} } = $key;
    }

    # is this enough ?
    $self->{sys}->{app} = $ref->{app};
    $self->{sys}->{app_version} = $ref->{app_version};
    $self->{sys}->{build_timestamp} = $ref->{build_timestamp};
    $self->{sys}->{id} = $ref->{root}->{id};
    $self->{sys}->{platform} = $ref->{platform};
    $self->{sys}->{product} = $ref->{product};
    $self->{sys}->{features} = $ref->{features}->{feature} ;

    # print "=== config/features ===\n", Dumper($self->{sys}->{features}), "\n";
    # print "=== config/features ===\n", Dumper($ref->{features}), "\n";
    $self->{cachetime}->{controls} = time;

    return $self;

}

sub check_feature {
    my $self = shift;
    my $id = shift;

    $self->load_conf() unless exists $self->{sys};

    $id = $featureids{$id} if exists $featureids{$id};

    return 1 if ( exists $self->{sys}->{features}->{$id} &&  $self->{sys}->{features}->{$id}->{isInstalled} eq "true" );

    return undef;
}

sub load_nodes {
    my $self = shift;	

    my $req;

    print "---load_nodes---\n" if $self->{debug} & 0x01;

    my $xurl = $self->{baseurl} . "/nodes";  # /nodes/devices
    print "xurl = $xurl\n", if $self->{debug} & 0x02;

    $req = HTTP::Request->new(GET => $xurl);
    $req->header('Accept' => 'text/xml,application/xhtml+xml,application/xml;q=0.9,*/*;q=0');

    my $res;

    $res = $self->{ua}->request($req);
    if (! $res->is_success) {
	$self->{error} = "FAIL : Devices : " .  $res->code . " : " . $res->status_line;
	print $self->{error}, "\n" if $self->{verbose};
	return undef;
    }
    my $keyattr = { 'node' => "+address",
		'group' => "+address",
		'folder' => "+address",
		'property' => "+id",
		'link' => "+content",
	    };
    my $ref = XMLin( ${$res->content_ref}, keyattr => $keyattr, ForceArray => [ 'property', 'folder' ]);

    print "=== Nodes ===\n", Dumper($ref), "\n" if $self->{debug} & 0x10;

    $res = undef;

    if (exists $self->{node_names} ) {
	delete  $self->{node_names} ;
	delete  $self->{node_x10} ;
	delete  $self->{node_x10_names};
	delete  $self->{scene_names} ;
    }


    $self->{folder} = $ref->{folder};
    $self->{nodes} = $ref->{node};
    $self->{scenes} = $ref->{group};

    for my $key (keys (  %{$ref->{ node }} ) ) {
	my $k = uc($ref->{node}->{$key}->{name});
	$self->{node_names}->{$k} = $key;
	if ( my $x = $self->x10_node($key)) {
	    $self->{node_x10}->{$x} = $key;
	    $self->{node_x10_names}->{$x} = $k;
	}

	if ( exists $ref->{node}->{$key}->{parent}
		&& $ref->{node}->{$key}->{parent}->{type} == 3 ) {
	    my $fa = $ref->{node}->{$key}->{parent}->{content};
	    $self->{folder}->{$fa}->{members} = () unless defined $self->{folder}->{$fa}->{members};
	    push( @{$self->{folder}->{$fa}->{members}}, $key);
	}

	#print "$ref->{node}->{$key}->{name} = $key\n";
	#- my $p = $ref->{node}->{$key}->{property};
	#- delete($ref->{node}->{$key}->{property} );
	#- $ref->{node}->{$key}->{property}->{ $p->{id} } = $p;

    }



    for my $key (keys (  %{$ref->{ group }} ) ) {
	my $k =  uc($ref->{group}->{$key}->{name});
	$self->{scene_names}->{$k} = $key;

	if ( exists $ref->{group}->{$key}->{parent}
		&& $ref->{group}->{$key}->{parent}->{type} == 3 ) {
	    my $fa = $ref->{group}->{$key}->{parent}->{content};
	    $self->{folder}->{$fa}->{members} = () unless defined $self->{folder}->{$fa}->{members};
	    push( @{$self->{folder}->{$fa}->{members}}, $key);
	}
    }

    for my $key ( keys (  %{$ref->{ folder }} ) ) {
	my $k =  uc($ref->{folder}->{$key}->{name});
	$self->{scene_names}->{$k} = $key;

	if ( exists $ref->{folder}->{$key}->{parent}
		&& $ref->{folder}->{$key}->{parent}->{type} == 3 ) {
	    my $fa = $ref->{folder}->{$key}->{parent}->{content};
	    $self->{folder}->{$fa}->{members} = () unless defined $self->{folder}->{$fa}->{members};
	    push( @{$self->{folder}->{$fa}->{members}}, $key);
	}
    }


    $self->{cachetime}->{nodes} = time;

    # print "=== scenes ===\n", Dumper($self->{scenes}), "\n";
    # print "=== Node_x10_names ===\n", Dumper($self->{node_x10_names}), "\n";
    # print "=== Node_names ===\n", Dumper($self->{node_names}), "\n";
    # print "=== Scene_names ===\n", Dumper($self->{scene_names}), "\n";

    return $self;
}

sub load_vars {
    my $self = shift;	
    my %args = @_;

    $self->{error} = undef;
    $self->{vars} = {};


    for my $t ( 1, 2 ) {
	print "---load_vars---  \n", if $self->{debug} & 0x1;;

	my $xurl = $self->{baseurl} . "/vars/get/$t";
	print "xurl = $xurl\n", if $self->{debug} & 0x02;

	my $req = HTTP::Request->new(GET => $xurl);

	$req->header('Accept' => 'text/xml,application/xhtml+xml,application/xml;q=0.9,*/*;q=0');
	my $res = $self->{ua}->request($req);
	if (! $res->is_success) {
	    $self->{error} = "FAIL : load_vars : " . $res->code . " : " . $res->status_line;
	    print $self->{error}, "\n" if $self->{verbose};
	    return undef;
	}
	my $ref = XMLin( ${$res->content_ref},  keyattr => { var => "+id" }, ForceArray => [ 'var' ]  );


	$xurl = $self->{baseurl} . "/vars/definitions/$t";
	print "xurl = $xurl\n", if $self->{debug} & 0x02;

	$req = HTTP::Request->new(GET => $xurl);
	$req->header('Accept' => 'text/xml,application/xhtml+xml,application/xml;q=0.9,*/*;q=0');
	$res = $self->{ua}->request($req);
	if (! $res->is_success) {
	    $self->{error} = "FAIL : load_vars : " . $res->code . " : " . $res->status_line;
	    print $self->{error}, "\n" if $self->{verbose};
	    return undef;
	}
	my $dref = XMLin( ${$res->content_ref},  keyattr => { e => "+id" }, ForceArray => [ 'e' ]  );

	if ( exists $ref->{var} ) {

	    $self->{vars}->{$t} = $ref->{var};

	    # merge var names with var list
	    for my $key (keys (  %{$dref->{ e }} ) ) {
		my $n = $dref->{e}->{$key}->{name};
		$self->{vars}->{$t}->{$key}->{name} = $n;
		$self->{vars_names}->{$t}->{$n} = $key;
	    }

#	    if ( exists $ref->{var}->{1} ) {
#		$self->{vars}->{$t} = $ref->{var};
#	    } elsif ( exists $ref->{var}->{id} ) {
#		# HACK ALERT, XML::Simple does not index by keyattr if there is only one element
#		$self->{vars}->{$t}->{ $ref->{var}->{id} } = $ref->{var};
#	    }

	}


    }

    if ( $self->{debug} & 0x08 ) {
	print "=== Vars ===\n", Dumper($self->{vars}), "\n";
	print "=== vars_names ===\n", Dumper($self->{vars_names}), "\n";
    }

    $self->{cachetime}->{vars} = time;

    return $self;
}

my %x10comm = (
    'alllightsoff' => 1,
    'status off' => 2,
    'on' => 3,
#    'preset dim' => 4,
    'alllightson' => 5,
#    'hail ack' => 6,
    'bright' => 7,
#    'status on'  => 8,
#    'extended code' => 9,
#    'status request' => 10,
    'off' => 11,
#    'preset dim' => 12,
    'alloff' => 13,
#    'Hail Req' => 14,
    'dim' => 15,
    'extended data' => 16,
);

sub x10_comm {
    my $self = shift;
    my $unit = uc(shift);
    my $command = lc(shift);

    my $xcomm;

    if ( $command =~ /^\d+$/ ) {
	$xcomm = $command;
    } else {
	if ( exists $x10comm{$command} ) {
	    $xcomm = $x10comm{$command};
	} else {
	    $self->{error} = "x10_comm : unknown command \"$command\" : $x10comm{$command} \n";
	    print $self->{error}, "\n" if $self->{verbose};
	    return undef;
	}
    }

    my $xurl =  $self->{baseurl} . "/X10/$unit/$xcomm";
    print "xurl = $xurl\n", if $self->{debug} & 0x02;

    my $req = HTTP::Request->new(GET => $xurl);
    $req->header('Accept' => 'text/xml,application/xhtml+xml,application/xml;q=0.9,*/*;q=0');
    my $res = $self->{ua}->request($req);
    if (! $res->is_success) {
	$self->{error} =  "FAIL : x10comm : " . $res->code . " : " . $res->status_line . "\n -- $xurl";
	print $self->{error}, "\n" if $self->{verbose};
	return undef;
    }

    return $res->code;

}

# converts "fake" INSTION X10 addr to X10 address
sub x10_node {
    my $self = shift;
    my $node = shift;

    if ( $node =~ /^FF (\S\S) (\S\S) \S$/ ) {
	    my $h = hex($1);
	    my $u = hex($2);
	    return undef if ( $h > 16 || $u > 16 );
	    $h = chr( $h + 64 );
	    return "$h$u";
    }

    return undef;
}

sub scene_names {
    my $self = shift;

    $self->load_nodes() unless exists  $self->{node_names};

    return keys %{$self->{scene_names}};
}

sub scene_addrs {
    my $self = shift;

    $self->load_nodes() unless exists  $self->{nodes};

    return keys %{$self->{scenes}};
}



sub node_names {
    my $self = shift;

    $self->load_nodes() unless exists  $self->{node_names};

    return keys %{$self->{node_names}};
}

sub node_addrs {
    my $self = shift;

    $self->load_nodes() unless exists  $self->{nodes};

    return keys %{$self->{nodes}};
}


### add arg to use cache or not
sub node_status {
    my $self = shift;
    my %args = @_;

    my $node = $self->get_node_id( $args{node} );

    unless ( $node ) {
	$self->{error} = "node unknown : $node";
	print $self->{error}, "\n" if $self->{verbose};
	return undef;
    }

    return $self->{nodes}->{$node}->{property}->{value};
}

### add arg to use cache or not
sub node_query {
    my $self = shift;
    #my %args = @_;
    my $arg = shift;

    $self->{error} = undef;

    my $node = $self->get_node_id( $arg );

    unless ( $node ) {
	$self->{error} = "node unknown : $arg";
	print $self->{error}, "\n" if $self->{verbose};
	return undef;
    }

    my $xurl =  $self->{baseurl} . "/nodes/$node";
    print "xurl = $xurl\n", if $self->{debug} & 0x02;

    my $req = HTTP::Request->new(GET => $xurl);
    $req->header('Accept' => 'text/xml,application/xhtml+xml,application/xml;q=0.9,*/*;q=0');

    my $res;

    $res = $self->{ua}->request($req);
    if (! $res->is_success) {
	$self->{error} = "FAIL : Node $node : " . $res->code . " : " . $res->status_line;
	print $self->{error}, "\n" if $self->{verbose};
	return undef;
    }
    my $keyattr = { 'node' => "+address",
		'group' => "+address",
		'folder' => "+address",
		'property' => "+id",
	    };
    #my $ref = XMLin( ${$res->content_ref}, keyattr => $keyattr);
    my $ref = XMLin( ${$res->content_ref}, keyattr => $keyattr, ForceArray => [ 'property', 'folder' ]);

    # print "=== QUERY ===\n", Dumper($ref), "\n";

    $self->{nodes}->{$node}->{property} = $ref->{properties}->{property};

    #print "=== \$ref ===\n", Dumper($ref->{properties}->{property}), "\n" if $self->{debug} & 0x10; 
    print "=== \$node ===\n", Dumper($self->{nodes}->{$node}), "\n" if $self->{debug} & 0x10; 

    return $ref->{properties}->{property};
}

sub get_node {
    my $self = shift;

    $self->{error} = undef;

    my $n=shift;
    my $node = $self->get_node_id( $n );

    # print "get_node $n : $node \n";

    unless ( $node ) {
	$self->{error} = "node unknown : $n";
	print $self->{error}, "\n" if $self->{verbose};
	return undef;
    }

    if ( exists $self->{nodes}->{$node} ) {
	return $self->{nodes}->{$node};
    }

    if ( exists $self->{scenes}->{$node}
	&& exists $self->{scenes}->{$node} ) {

	return $self->{scenes}->{$node};
    }

    return undef;
}


# same as node_comm but node and command are not checked
# thus the node and command list do not need to be preloaded and parced
# thus saving time if your just trying to send a only a quick command
sub node_comm_simple {
    my $self = shift;

    return $self;

}

sub node_comm {
    my $self = shift;

#    my %args = @_;
#    my $node = $self->get_node_id( $args{node} );
#    my $comm = $self->get_comm_id( $args{comm} );
#    my @param = $args{param};

    my $n=shift;
    my $node = $self->get_node_id( $n );
    my $c=shift;
    my $comm = $self->get_comm_id( $c );
    my @param = @_;

    $self->{error} = undef;

    unless ( $node ) {
	$self->{error} = "node unknown : $n";
	print $self->{error}, "\n" if $self->{verbose};
	return undef;
    }

    unless ( $comm ) {
	$self->{error} = "command unknown : $c";
	print $self->{error}, "\n" if $self->{verbose};
	return undef;
    }

    my $xurl =  $self->{baseurl} . "/nodes/$node/cmd/$comm";

    if ( $param[0] =~ /(\d+)%/ ) {
	my $p = $1;
	$param[0] = 255 * ( $p / 100 );
	print "$1 -> $param[0]\n" if $self->{debug} & 0x02;
    }

    for my $p ( @param ) {
	$xurl .= "/$p";
    }

    # print "SEND_COMM : node = $node : comm = $comm : param #= ", scalar(@param), "\n\txurl=$xurl\n";

    print "xurl = $xurl\n" if $self->{debug} & 0x02;

    my $req = HTTP::Request->new(GET => $xurl);
    $req->header('Accept' => 'text/xml,application/xhtml+xml,application/xml;q=0.9,*/*;q=0');
    my $res = $self->{ua}->request($req);
    if (! $res->is_success && $comm ne "BEEP") {
	$self->{error} = "Invalid node command : " .  $res->code .  " : " .  $xurl;
	print $self->{error}, "\n" if $self->{verbose};
	return undef;
    }
    return $self;
}

# prog func

sub prog_names {
    my $self = shift;

    $self->load_prog() unless exists  $self->{prog_names};

    return keys %{$self->{prog_names}};
}


sub prog_status {
    my $self = shift;
    my $arg = shift;

    $self->{error} = undef;

    my $prog = $self->get_prog_id( $arg );

    unless ( $prog ) {
	$self->{error} = "program unknown : $arg\n";
	print $self->{error}, "\n";
	return undef;
    }

    return $self->{programs}->{$prog}->{status};
}

sub prog_query {
    my $self = shift;
    # my %args = @_;
    my $arg = shift;

    $self->{error} = undef;

    my $prog = $self->get_prog_id( $arg );

    unless ( $prog ) {
	$self->{error} = "prog unknown : $arg";
	print $self->{error}, "\n";
	return undef;
    }

    my $xurl =  $self->{baseurl} . "/programs/$prog";
    print "xurl = $xurl\n", if $self->{debug} & 0x02;

    my $req = HTTP::Request->new(GET => $xurl);
    $req->header('Accept' => 'text/xml,application/xhtml+xml,application/xml;q=0.9,*/*;q=0');
    my $res = $self->{ua}->request($req);
    if (! $res->is_success) {
	$self->{error} = "FAIL : prog_query : " . $res->code . " : " . $res->status_line;
	print $self->{error}, "\n";
	return undef;
    }
    my $ref = XMLin( ${$res->content_ref} );

    #print "=== Prog ===\n", Dumper($ref), "\n";

    $self->{programs}->{$prog} = $ref->{program};

    #print "=== Programs ===\n", Dumper($self->{programs}), "\n";

    return $ref->{program};
}

# run|runThen|runElse|stop|enable|disable|enableRunAtStartup|disableRunAtStartup
sub prog_comm {
    my $self = shift;
    my %args = @_;

    $self->{error} = undef;

    ###
    my $prog = $self->get_prog_id( $args{prog} );
    my $comm = $args{comm};

    unless ( $prog ) {
	print "prog unknown : $prog\n";
	return undef;
    }

    print "PROG_COMM : prog = $args{prog} / $prog : comm = $args{comm}\n";

    my $xurl =  $self->{baseurl} . "/programs/$prog/$comm";
    print "xurl = $xurl\n", if $self->{debug} & 0x02;

    my $req = HTTP::Request->new(GET => $xurl);
    $req->header('Accept' => 'text/xml,application/xhtml+xml,application/xml;q=0.9,*/*;q=0');
    my $res = $self->{ua}->request($req);
    if (! $res->is_success) {
	$self->{error} = "Invalid program or command : " .  $res->code .  " : " .  $xurl;
	print $self->{error}, "\n";
	return undef;
    }

    # print "\tURL = $xurl : code = ", $res->code, "\n";

    return $self;
}



sub nres_comm {
    my $self = shift;
    my $res_name = shift;

    $self->{error} = undef;

    my $res_id = $self->get_res_id( $res_name );

    print "NRES_COMM : res_name = $res_name : res_id = $res_id\n";

    unless ( $res_id ) {
	print "network resource id unknown : $res_name\n";
	return undef;
    }

    my $xurl = $self->{baseurl} . "/networking/resources/$res_id";
    print "xurl = $xurl\n", if $self->{debug} & 0x02;

    my $req = HTTP::Request->new(GET => $xurl);
    $req->header('Accept' => 'text/xml,application/xhtml+xml,application/xml;q=0.9,*/*;q=0');
    my $res = $self->{ua}->request($req);
    if (! $res->is_success) { # this should never happen
	$self->{error} = "Invalid network resource id or command : " .  $res->code .  " : " .  $xurl;
	print $self->{error}, "\n" if $self->{verbose};
	return undef;
    }

    return $self;
}


sub wol_comm {
    my $self = shift;
    my $wol_name = shift;

    $self->{error} = undef;

    my $wol_id = $self->get_wol_id( $wol_name );

    # print "WOL_COMM : wol_name = $wol_name : wol_id = $wol_id\n";

    unless ( $wol_id ) {
	$self->{error} = "WOL ID unknown : $wol_name";
	return undef;
    }

    my $xurl =  $self->{baseurl} . "/networking/wol/$wol_id";
    print "xurl = $xurl\n", if $self->{debug} & 0x02;

    my $req = HTTP::Request->new(GET => $xurl);
    $req->header('Accept' => 'text/xml,application/xhtml+xml,application/xml;q=0.9,*/*;q=0');
    my $res = $self->{ua}->request($req);
    if (! $res->is_success) { # this should never happen
	$self->{error} = "Invalid WOL id or command : " .  $res->code .  " : " .  $xurl;
	print $self->{error}, "\n" if $self->{verbose};
	return undef;
    }

    return $self->{networking}->{wol}->{$wol_id};
}

sub wol_query {
    my $self = shift;

    if (  ! exists  $self->{networking} ) {
	return undef unless $self->load_netw();
    }

    return( $self->{networking}->{wol} );
}

sub var_query {
    my $self = shift;
    my $var_name = shift;

    if (  ! exists  $self->{vars} ) {
	return undef unless $self->load_vars();
    }

    return( $self->{vars} );

#    if ( defined $var_name ) {
#	my ($var_type, $var_id) = $self->get_var_id( $var_name );
#	my $h = {};
#	$h->{$var_type}->{$var_id} = $self->{vars}->{$var_type}->{$var_id};
#	return( $h ) ;
#    } else {
#	return( $self->{vars} );
#    }
}

sub var_get {
    my $self = shift;
    my $var_name = shift;

    if (  ! exists  $self->{vars} ) {
	return undef unless $self->load_vars();
    }

    my ($var_type, $var_id) = $self->get_var_id( $var_name );

    unless ($var_type) { 
	$self->{error} = "var unknown : $var_name";
	print $self->{error}, "\n" if $self->{verbose};
	return undef;
    }

# 	print "var_get : $var_type $var_id  $self->{vars}->{$var_type}->{$var_id}->{val}\n";
	return $self->{vars}->{$var_type}->{$var_id}->{val};
}

sub var_set {
    my $self = shift;
    my ($n, $v, $i) = @_;

    print "---var_set : $n $v --- \n"; # if $self->{debug} & 0x01;

    if (  ! exists  $self->{vars} ) {
	return undef unless $self->load_vars();
    }

    my ($var_type, $var_id) = $self->get_var_id( $n );

    unless ($var_type) { 
	$self->{error} = "var unknown : $n";
	print $self->{error}, "\n" if $self->{verbose};
	return undef;
    }

    my $xurl =  $self->{baseurl};

    if ( $i ) {
   	$xurl .= "/vars/init/$var_type/$var_id/$v";
    } else {
   	$xurl .= "/vars/set/$var_type/$var_id/$v";
    }
    print "xurl = $xurl\n", if $self->{debug} & 0x02;
    my $req = HTTP::Request->new(GET => $xurl);
    $req->header('Accept' => 'text/xml,application/xhtml+xml,application/xml;q=0.9,*/*;q=0');
    my $res = $self->{ua}->request($req);
    if (! $res->is_success ) {
	$self->{error} = "Error set var : " .  $res->code .  " : " .  $xurl;
	print $self->{error}, "\n" if $self->{verbose};
	return undef;
    }

    $self->{vars}->{$var_type}->{$var_id}->{val} = $v;

    return $self;
}




sub clim_query {
    my $self = shift;

    if (  ! exists  $self->{climate} ) {
	return undef unless $self->load_clim();
    }

    return( $self->{climate} );
}


### LOG

sub log_reset {
    my $self = shift;
    my %args = @_;

    return $self->log_query( reset => 1, error => $args{error} );
}

sub log_query {
    my $self = shift;
    my %args = @_;
    my $logs = [ ];
    my $a = [] ;

    $self->{error} = undef;
    $self->load_log_ids() unless exists  $self->{logactors};


    my $xurl = $self->{baseurl} . "/log" ;

    if ( $args{error} ) {
	$xurl .= "/error";
    }

    if ( $args{reset} ) {
	$xurl .= "?reset=true";
    }

    print " xurl = $xurl\n" if $self->{debug} & 0x02;

    my $req = HTTP::Request->new(GET => $xurl);
    $req->header('Accept' => 'text/xml,application/xhtml+xml,application/xml;q=0.9,*/*;q=0');
    my $res = $self->{ua}->request($req);

    if (! $res->is_success) {
	return $a;

    }

    unless ( $args{reset} ) {
	@{$a} = split(/[\n\r]/, ${$res->content_ref});
    }

#    print "\@a = ", scalar(@{$a}), "\n";
    return $a;
}


my %log_uid = (
    0 => "SYSTEM_USER",
    1 => "SYSTEM_DRIVER_USER",
    2 => "WEB_USER",
    3 => "SCHEDULER_USER",
    4 => "D2D_USER",
    5 => "ELK_USER",
    6 => "SEP_DEVICE_UMETER_USER",
    7 => "SEP_DEVICE_UPRICE_USER",
    8 => "SEP_DEVICE_UMSG_USER",
    9 => "SEP_DEVICE_UDR_USER",
    10 => "GAS_METER_USER",
);


my %DEV_CAT = (
    '0' => 'CONTROLLER',
    '1' => 'DIM_LIGHT_CONTROL',
    '10' => 'APPLIANCE_CONTROL',
    '11' => 'PLUMBING',
    '113' => 'X10',
    '12' => 'COMMUNICATION',
    '127' => 'VIRTUAL',
    '13' => 'COMPUTER_CONTROL',
    '14' => 'BIPHY_LAMPLINC_B2457D2',
    '15' => 'ACCESS_CONTROL',
    '16' => 'SECURITY_HEALTH_SAFETY',
    '17' => 'SURVEILLANCE',
    '18' => 'AUTOMOTIVE',
    '19' => 'PET_CARE',
    '2' => 'SWITCHED_LIGHT_CONTROL',
    '20' => 'TOYS',
    '21' => 'TIME_KEEPING',
    '22' => 'HOLIDAY',
    '222' => 'SWITCH_LIGHT_CONTROL',
    '254' => 'UNKNOWN',
    '3' => 'NETWORK_BRIDGE',
    '4' => 'IRRIGATION_CONTROL',
    '5' => 'CLIMATE_CONTROL',
    '6' => 'POOL_CONTROL',
    '7' => 'SENSOR_ACTUATOR',
    '8' => 'HOME_ENTERTAINMENT',
    '9' => 'ENERGY_MANAGEMENT'
);

my %DEV_SCAT = (
    'ACCESS_CONTROL' => {
      '0' => 'WEILAND_CENTRAL_DRIVE_CONTROLLER',
      '1' => 'WEILAND_SECONDARY_CENTRAL_DRIVE',
      '2' => 'WEILAND_ASSIST_DRIVE',
      '3' => 'WEILAND_ELEVATION_DRIVE',
      '6' => 'MORNING_LINC'
    },
    'CLIMATE_CONTROL' => {
      '0' => 'BROAN_SMSC080_EXHAUST_FAN',
      '1' => 'COMPACTA_EZTHERM',
      '2' => 'BROAN_SMSC110_EXHAUST_FAN',
      '3' => 'INSTEON_THERMOSTAT_ADAPTER',
      '4' => 'COMPACTA_EZTHERMX',
      '5' => 'BROAN_VENMAR_BEST'
    },
    'CONTROLLER' => {
      '0' => 'CONTROLINC_2430',
      '10' => 'POOLUX_LCD_CONTROLLER',
      '11' => 'ACCESSPOINT',
      '12' => 'IES_COLOR_TOUCHSCREEN',
      '5' => 'ICON_REMOTELINC_2843',
      '6' => 'ICON_TABLETOP_2830',
      '9' => 'SIGNALINC_2442'
    },
    'DIM_LIGHT_CONTROL' => {
      '0' => 'LAMPLINC_V2_2456D3',
      '1' => 'SWITCHLINC_V2_DIMMER_2476D',
      '10' => 'ICON_INWALL_CONTROLLER_2886D',
      '12' => 'KEYPADLINC_DIMMER_2486DWH8',
      '13' => 'SOCKETLINC_2454D',
      '19' => 'ICON_SWITCHLINC_DIMMER_BELL_CANADA',
      '2' => 'INLINE_DIMMABLE',
      '23' => 'TOGGLELINC_DIMMER_2466D',
      '24' => 'COMPANION_SWITCH_2474D',
      '25' => 'SWITCHLINC_DIMMER_W_SENSE_2476D',
      '26' => 'INLINELINC_DIMMER_2475D',
      '27' => 'KEYPAD_LINC_DIMMER_2486D_6',
      '28' => 'KEYPAD_LINC_DIMMER_2486D_8',
      '29' => 'SWITCH_LINC_DIMMER_2476DH',
      '3' => 'ICON_SWITCH_DIMMER_2876D3',
      '30' => 'ICON_SWITCH_DIMMER_2876DB',
      '31' => 'TOGGLELINC_DIMMER_2466D_2',
      '32' => 'SWITCHLINC_DIMMER_2477D',
      '34' => 'LAMPLINC_2_PIN_DIMMER_2457D2X',
      '4' => 'SWITCHLINK_V2_DIMMER_2476DH',
      '5' => 'KEYPADLINC_TIMER_2484DWH8',
      '6' => 'LAMPLINC_2_PIN',
      '7' => 'ICON_LAMPLINC_V2_2_PIN_2456D2',
      '9' => 'KEYPADLINC_DIMMER_2486D'
    },
    'ENERGY_MANAGEMENT' => {
      '0' => 'COMPACTA_EZENERGY',
      '1' => 'ONSITE_PRO_LEAK_DETECTOR',
      '10' => 'DUAL_BAND_NO_RELAY_240V_2477SA1',
      '11' => 'DUAL_BAND_NC_RELAY_240V_2477SA2',
      '2' => 'ONSITE_PRO_CONTROL_VALVE',
      '3' => 'TED_5000_SINGLE_PHASE',
      '4' => 'TED_5000_GATEWAY_USB',
      '5' => 'TED_5000_GATEWAY_ETHERNET',
      '6' => 'TED_5000_THREE_PHASE',
      '7' => 'IMETER_SOLO'
    },
    'IRRIGATION_CONTROL' => {
      '0' => 'COMPACTA_EZRAIN_SPRINKLER_CONTROLLER'
    },
    'NETWORK_BRIDGE' => {
      '1' => 'POWERLINC_SERIAL_2414S',
      '11' => 'POWERLINC_MODEM_USB',
      '13' => 'EZX10RF',
      '15' => 'EZX10IR',
      '16' => 'SMARTlINC_2412N_CENTRAL_CONTROLLER',
      '2' => 'POWERLINC_USB_2414U',
      '3' => 'ICON_POWERLINC_SERIAL_2814S',
      '4' => 'ICON_POWERLINC_USB_2814U',
      '5' => 'POWERLINE_MODEM',
      '6' => 'IRLINC',
      '7' => 'IRLINC_TX'
    },
    'POOL_CONTROL' => {
      '0' => 'COMPACTA_EZPOOL',
      '222' => 'COMPACTA_EZ_POOL'
    },
    'SECURITY_HEALTH_SAFETY' => {
      '0' => 'FIRST_ALERT_ONELINK',
      '1' => 'MOTION_SENSOR_2420M',
      '2' => 'TRIGGER_LINC_2421'
    },
    'SENSOR_ACTUATOR' => {
      '0' => 'IO_LINC_2450',
      '1' => 'COMPACTA_EZSENSE',
      '2' => 'COMPACTA_EZIO_8T',
      '3' => 'COMPACTA_EZIO',
      '4' => 'COMPACTA_EZIO_8',
      '5' => 'COMPACTA_EZSNS_RF',
      '6' => 'COMPACTA_EZISNS_RF',
      '7' => 'COMPACTA_EZIO_6I',
      '8' => 'COMPACTA_EZIO_4O',
      '9' => 'SYNCHRO_LINC'
    },
    'SWITCHED_LIGHT_CONTROL' => {
      '10' => 'SWITCHLINC_RELAY_2476S',
      '11' => 'ICON_ON_OFF_SWITCH_2876S',
      '12' => 'ICON_APPLIANCE_ADAPTER_2856S3',
      '13' => 'TOGGLELINC_RELAY_2466S',
      '14' => 'SWITCHLINC_RELAY_COUNTDOWN_TIMER_2476ST',
      '15' => 'KEYPADLINC_RELAY_2486S',
      '16' => 'INLINE_RELAY',
      '17' => 'EZSWITCH_30',
      '18' => 'COMPANION_SWITCH_2474S',
      '19' => 'ICON_SWTICHLINC_RELAY_BELL_CANADA',
      '20' => 'INLINE_RELAY_WITH_SENSE',
      '21' => 'SWITCHLINC_RELAY_W_SENSE_2476S',
      '22' => 'ICON_RELAY_2876SB',
      '23' => 'ICON_APPLIANCELINC_2856S3B',
      '24' => 'SWITCHLINC_RELAY_2494S220',
      '25' => 'SWITCHLINC_RELAY_2494S220_B',
      '26' => 'TOGGLELINC_RELAY_2466S_2',
      '28' => 'SWITCHLINC_RELAY_REMOTE_CONTROL_2476S',
      '5' => 'KEYPADLINC_RELAY_2486SWH8',
      '6' => 'APPLIANCELINC_OUTDOOR_2456S3E',
      '7' => 'TIMERLINC_2456S3T',
      '8' => 'OUTLETLINC_2473',
      '9' => 'APPLIANCELINC_2456S3'
    },
    'WINDOWS_COVERING' => {
      '0' => 'SOMFY_DRAPE_CONTROLLER_RF'
    },
    'X10' => {
      '1' => 'X10',
      '126' => 'VIRTUAL',
      '2' => 'A10',
      '254' => 'UNKNOWN'
    }
);



sub node_type {
    my $self = shift;
    my $type = shift;
    my $devcat;

    $self->load_node_types() unless exists  $self->{nodeCategory};

    my @a = split(/\./, $type);

    if ( ! exists $self->{nodeCategory}->{$a[0]}) {
	return ( "UNKNOWN", "UNKNOWN", $a[2]);
    }

    $a[2] = "v.$a[2]";

    $devcat = $self->{nodeCategory}->{$a[0]}->{name};
    #$devcat=$DEV_CAT{$a[0]};

    unless ( exists $self->{nodeFam}->{$a[0]} && $self->{nodeFam}->{$a[0]}->{nodeSubCategory}->{$a[1]}) {
	 return ( "$devcat", "UNKNOWN", $a[2]);
    }

    my $subcat = $self->{nodeFam}->{$a[0]}->{nodeSubCategory}->{$a[1]}->{name};
    $subcat =~ s/^DEV_SCAT_//;

    return ( "$devcat", $subcat, $a[2]);

}

sub scene_nodes {
    my $self = shift;
    my $s = shift;
    my $scene;

    $self->{error} = undef;

    unless (  $scene = $self->get_node_id( $s ) ) {
	$self->{error} = "Unknown scene : $s";
	return undef;
    }

    if  ( exists $self->{scenes}->{$scene} ) {
	return keys ( %{$self->{scenes}->{$scene}->{members}->{link}});
    }

    return ();
}

sub node_scenes {
    my $self = shift;
    my $n = shift;
    my $node;

    $self->{error} = undef;

    unless (  $node = $self->get_node_id( $n ) ) {
	$self->{error} = "Unknown node : $n";
	return undef;
    }

    # print "node_scenes : $n : $node\n";
    $self->gen_node_scenes unless defined $self->{node_scenes};

    if  ( exists $self->{node_scenes}->{$node} ) {
	return @{$self->{node_scenes}->{$node}};
    } 

    return ();
}

sub gen_node_scenes {
    my $self = shift;

    delete $self->{node_scenes} if defined $self->{node_scenes};

    for my $sc (keys (  %{$self->{ scenes }} ) ) {
	for my $mem ( keys ( %{$self->{scenes}->{$sc}->{members}->{link}} ) ) {
	    $self->{node_scenes}->{$mem} = ( ) unless exists $self->{node_scenes}->{$mem};
	    push( @{$self->{node_scenes}->{$mem}}, $sc);
	}
    }

    return;
}



my %log_types = (
    1 => "SYSTEM_STARTUP",
    2 => "SYSTEM_SHUTDOWN",
    3 => "WARNING",
    4 => "INFO",
    5 => "LOG",
    6 => "UD_SEP_SUBSYS_STARTUP",
    -1 => "REQUEST_FAILED_ERROR",
    -2 => "DEVICE_COMMUNICATION_ERROR",
    -3 => "DEVICE_RETURNED_INVALID_NODE",
    -4 => "DEVICE_RETURNED_INVALID_ADDRESS",
    -5 => "ERROR_LOGGER_STARTUP",
    -10 => "MAIN_HAML_DRIVER_NOT_FOUND",
    -20 => "MAIN_LOCAL_DEVICE_BLANK",
    -100 => "SYSTEM_NO_NETWORK_CONNECTION",
    -101 => "SYSTEM_WEBSERVER_SELECT_FAILED",
    -500 => "HAML_DRIVER_LISTENER_NOT_REGISTERED",
    -1000 => "HAML_PARSER_UNDEFINED_ELEMENT",
    -1001 => "HAML_PARSER_ONDATA",
    -5001 => "UPNP_DRIVER_NO_DEVICES_CONFIGURED",
    -5002 => "UPNP_DRIVER_SERIAL_READER_FAILED",
    -5003 => "UPNP_DRIVER_MAX_DEVICES",
    -5004 => "UPNP_SERVICE_TYPE_SEARCH_NS",
    -5005 => "UPNP_SUBSCRIPTION_NOT_FOUND_FOR_RENEWAL",
    -5006 => "UPNP_SUBSCRIPTION_NOT_FOUND_FOR_CANCELATION",
    -5007 => "UPNP_INVALID_SUBSCRIPTION_URL",
    -5008 => "UPNP_INVALID_SUBSCRIPTION_CALLBACK",
    -5009 => "UPNP_MAX_SUBSCRIBERS",
    -5010 => "UPNP_SUBSCRIBER_TCP_CONNECT_FAILURE",
    -5011 => "PROCESS_DEVICE_STATE_CHANGE_SID_NOT_FOUND",
    -5012 => "UPNP_SUBSCRIBER_NOREPLY_TO_EVENT_1",
    -5013 => "UPNP_SUBSCRIBER_NOREPLY_TO_EVENT_2",
    -5014 => "UPNP_SUBSCRIBER_NOREPLY_TO_EVENT_3",
    -5015 => "UPNP_CONTROL_MALFORMED_SOAP_REQUEST_1",
    -5016 => "UPNP_CONTROL_MALFORMED_SOAP_REQUEST_2",
    -6000 => "OS_DUPLICATE_TASK_PRIORITY",
    -6001 => "OS_OPEN_SERIAL_FAILED",
    -7020 => "D2D_PARSER_ERROR",
    -7029 => "NOTIFICATIONS_MAIL_TO_ADDRESS_REQUIRED",
    -7030 => "NOTIFICATIONS_SEND_MAIL_FAILED",
    -7050 => "D2D_EXPECTED_D2D_TAG",
    -7051 => "D2D_UNEXPECTED_TAG_IN_SENSE",
    -7052 => "D2D_UNEXPECTED_TAG_IN_CONDITION",
    -7501 => "DIAG_PARSER_ERROR",
    -7601 => "LINK_PARSER_ERROR",
    -10100 => "PNP_SECURITY_NOT_VERIFIED",
    -10001 => "SSL_DECODING_LENGTHS_FAILED",
    -10002 => "SSL_DECODING_PMOD_FAILED",
    -10003 => "SSL_DECODING_PEXP_FAILED",
    -10004 => "SSL_DECODING_PRI_EXP_FAILED",
    -10005 => "SSL_DECODING_PRI_P_FAILED",
    -10006 => "SSL_DECODING_PRI_Q_FAILED",
    -10007 => "SSL_DECODING_PRI_X1_FAILED",
    -10008 => "SSL_DECODING_PRI_X2_FAILED",
    -10009 => "SSL_DECODING_COEFF_FAILED",
    -10010 => "SSL_DECODING_CERT_FAILED",
    -10011 => "SSL_REQUEST_NOT_AUTHENTICATED",
    -10026 => "SECURE_SESSION_DOES_NOT_EXIST",
    -10027 => "SECURE_SESSIONS_EXHAUSTED",
    -10101 => "AUTHENTICATION_UNSUPPORTED_UID_LEN",
    -10102 => "AUTHENTICATION_UNSUPPORTED_PWD_LEN",
    -10103 => "AUTHENTICATION_USER_ID_DOES_NOT_EXIST",
    -10104 => "AUTHENTICATION_USER_ID_PWD_NOT_PRESENT",
    -10105 => "AUTHENTICATION_WRONG_PASSWORD",
    -10106 => "AUTHENTICATION_FAILED",
    -10107 => "HTTP_AUTH_DECODING_FAILED",
    -11000 => "SECURITY_INITIALIZATION_FAILED",
    -12000 => "TIMED_OUT_WAITING_FOR_CRITICAL_SECION",
    -12001 => "ERROR_LEAVING_CRITICAL_SECTION_NOT_OWNED",
    -13000 => "CONTENT_LEN_NOT_EQUAL_TO_HEADER_CONTENT_LEN",
    -14001  => "XML_MALFORMED_TAG",
    -14002 => "XML_MALFORMED_END_TAG",
    -14003  => "XML_NO_START_TAG",
    -14004  => "XML_NO_TAG_NAME",
    -14005  => "XML_START_END_NAME_MISMATCH",
    -20000 => "MALFORMED_UPNP_HEADERS",
    -50000 => "MAIL_SERVER_CONNECT_ERROR",
    -50001 => "SMTP_SERVER_FAILURE",
    -50010 => "MAIL_SERVER_DNS_ERROR",
    -50011 => "MAIL_MAX_FROM_LEN",
    -50012 => "MAIL_MAX_SUBJECT_LEN",
    -50013 => "MAIL_MAX_TO_LEN",
    -60000 => "NTP_CONFIG_SERVER_NO_HOST_PARAM",
    -60001 => "NTP_CONFIG_SERVER_ADDRESS_RESOLUTION_FAILED",
    -60002 => "NTP_CONFIG_SERVER_NO_INTERVAL_PARAM",
    -60006 => "NTP_SERVER_NOT_RESPONDING",
    -60007 => "NTP_SERVER_CONNECT_ERROR",
    -70000 => "OUT_OF_MEMORY",
    -80000 => "IGD_FAILED_PARSING_DESCRIPTION_URL",
    -80001 => "IGD_FAILED_RETRIEVING_DESCRIPTION_FILE",
    -80002 => "IGD_FAILED_RETRIEVING_URL_BASE",
    -80003 => "IGD_FAILED_PARSING_URL_BASE",
    -80004 => "IGD_FAILED_RETRIEVING_WAN_CONNECTION_DEVICE",
    -80005 => "IGD_FAILED_RETRIEVING_CONTROL_URL",
    -80006 => "IGD_FAILED_PARSING_CONTROL_URL",
    -80007 => "IGD_FAILED_RETRIEVING_EXTERNAL_IP",
    -80008 => "IGD_NO_RESPONSE_FROM_GATEWAY",
    -80009 => "IGD_FAILED_STRIPPING_HTTP_HEADERS",
    -80010 => "IGD_FAILED_DELETING_PORT_FORWARD_MAP",
    -80011 => "IGD_FAILED_ADDING_PORT_FORWARD_MAP",
    -80012 => "IGD_FAILED_GETTING_SPECIFIC_ENTRY",
    -90001 => "CRC_INVALID_ORDER",
    -90002 => "CRC_INVALID_POLYNOM",
    -90003 => "CRC_INVALID_CRC_INIT",
    -90004 => "CRC_INVALID_CRC_XOR",
    -100000 => "LOGGER_DIRECTORY_CREATION_FAILED",
    -100001 => "LOGGER_SD_IS_NOT_INSTALLED",
    -100002 => "LOGGER_LOG_FILE_OPEN_FAILED",
    -110000 => "FILE_TO_STRING_OPEN_FAILED",
    -110001 => "FILE_TO_STRING_MEM_ALLOC_FAILED",
    -110002 => "SD_DRIVE_FORMAT_FAILED_1",
    -110003 => "SD_DRIVE_FORMAT_FAILED_2",
    -110004 => "SD_DRIVE_MOUNT_FAILED_1",
    -110005 => "SD_DRIVE_MOUNT_FAILED_2",
    -110006 => "SEND_FILE_OPEN_FAILED",
    -110007 => "SEND_FILE_READ_FAILED",
    -110008 => "RECEIVE_FILE_WRITE_FAILED",
    -110009 => "RECEIVE_FILE_OPEN_FAILED",
    -110010 => "SD_DRIVE_DIRECTORY_CREATION_FAILED",
    -110011 => "SD_DRIVE_CONFIG_FILE_OPEN_WRITE_FAILED",
    -110012 => "SD_DRIVE_CONFIG_FILE_OPEN_READ_FAILED",
    -110013 => "SD_DRIVE_CONFIG_WRITE_FAILED",
    -110014 => "SD_DRIVE_CONFIG_READ_FAILED",
    -110015 => "STRING_TO_FILE_OPEN_FAILED",
    -110016 => "STRING_TO_FILE_WRITE_FAILED",
    -110017 => "FILE_TO_STRING_READ_FAILED",
    -110018 => "REMOVE_FILE_FAILED",
    -110019 => "REMOVE_DIR_FAILED",
    -110020 => "FLUSH_FILE_FAILED",
    -110021 => "CLOSE_FILE_FAILED",
    -110022 => "OPEN_FILE_FAILED",
    -110023 => "FLUSH_FILE_SYSTEM_FAILED",
    -110024 => "FILESYSTEM_INIT_FAILED",
    -110025 => "FILESYSTEM_CRIT_FAILED",
    -120000 => "FIRMWARE_UPDATE_OPEN_FILE_FAILED",
    -120001 => "FIRMWARE_UPDATE_HEADER_READ_FAILED",
    -120002 => "FIRMWARE_UPDATE_CHECKSUM_FAILED",
    -120003 => "FIRMWARE_UPDATE_MALLOC_FAILED",
    -120004 => "FIRMWARE_UPDATE_DATA_READ_FAILED",
    -130000 => "ELK_CONFIG_PARSER_ERROR",
    -140000 => "HTTP_CLIENT_DNS_ERROR",
    -140001 => "HTTP_CLIENT_BASE64_ENCRYPTION_FAILED",
    -140002 => "HTTP_CLIENT_CONNECTION_TIMED_OUT",
    -140003 => "HTTP_CLIENT_WRITE_HEADER_FAILED",
    -140004 => "HTTP_CLIENT_WRITE_BODY_FAILED",
    -140005 => "HTTP_CLIENT_READ_RESPONSE_FAILED",
    -140006 => "HTTP_CLIENT_HEADER_NO_STATUS",
    -140007 => "HTTP_CLIENT_RESOURCE_MOVED",
    -140008 => "HTTP_CLIENT_REQUEST_FAILED",
    -140009 => "HTTP_CLIENT_NO_NETWORK",
    -150000 => "TCP_CLIENT_WRITE_FAILED",
    -150100 => "UDP_CLIENT_DNS_ERROR",
    -160000 => "PROTOCOL_READER_READ_ERROR",
    -160001 => "PROTOCOL_READER_BUFFER_OVERFLOW",
    -160002 => "PROTOCOL_READER_REOPEN_ERROR",
    -170000 => "WEB_MODULE_NO_FREE_SPACE",
    -170001 => "SYSTEM_ACCESS_LOG",
    -180000 => "SEP_NETWORK_SCAN_ERROR",
    -180001 => "SEP_NETWORK_KEY_EST_ERROR",
    -180002 => "SEP_NETWORK_DISCOVERY_ERROR",
    -180003 => "SEP_NETWORK_SYNCH_ERROR",
    -180004 => "SEP_MODULE_RESET_ERROR",
    -180005 => "SEP_MODULE_INVALID_CALL_ERROR",
    -180006 => "SEP_MODULE_UNKNOWN_ERROR",
    -190001 => "UDERR_ISY_API_NO_SPACE",
    -190002 => "UDERR_ISY_API_INVALID_8_3_FILENAME",
    -190003 => "UDERR_ISY_API_INVALID_PGM_FILENAME",
    -190004 => "UDERR_ISY_API_INCORRECT_PGM_KEY",
    -190005 => "UDERR_ISY_API_INVALID_PGM_URL_SEARCH_STRING",
    -200000 => "DEVICE_DRIVER_ERROR_MSG",
    -210001 => "CALL_HOME_PORTAL_NO_FD",
);

sub log_type {
    my $self = shift;
    my $t = shift;

    return $log_types{$t} if exists $log_types{$t} ;

    return undef;
}

sub get_prog_id {
    my $self = shift;
    my $prog = uc( shift );

#print "get_prog_id : $prog\n";

    $self->load_prog() unless exists  $self->{prog_names};

    if ( exists ( $self->{prog_names}->{$prog} ) ) {
	return ( $self->{prog_names}->{$prog} );
    }

    if ( exists ( $self->{programs}->{$prog} ) ) {
	return ( $prog );
    }

    return undef;
}


sub get_wol_id {
    my $self = shift;
    my $wol_id =  shift;

#print "get_wol_id : $wol_id\n";

    $self->load_netw() unless exists  $self->{networking};

    if ( exists ( $self->{networking}->{wol_names}->{$wol_id} ) ) {
	return (  $self->{networking}->{wol_names}->{$wol_id} );
    }

    if ( exists ( $self->{networking}->{wol}->{$wol_id} ) ) {
	return ( $wol_id );
    }

    return undef;
}


sub get_var_id {
    my $self = shift;
    my $var_id = shift;

#print "get_var_id : $var_id\n";

    $self->load_vars() unless exists  $self->{vars};

    if ( $var_id =~ /(\d+):(\d+)/ ) {
	if ( exists ( $self->{vars}->{$1}->{$2} ) ) {
	    return ($1, $2);
	}
    }

    for my $t ( 1, 2 ) {

	if ( exists ( $self->{vars_names}->{$t}->{$var_id} ) ) {
	    return ( $t, $self->{vars_names}->{$t}->{$var_id} );
	}

	if ( exists ( $self->{vars}->{$t}->{$var_id} ) ) {
	    return ( $t, $var_id );
	}
    }

    return undef;
}

sub get_x10_addr {
    my $self = shift;
    my $x = uc( shift );

    if ( defined $self->{node_x10}->{$x} ) {
	return $self->{node_x10}->{$x};
    }

    return undef;
}


sub get_x10_name {
    my $self = shift;
    my $x = uc( shift );

    $self->load_nodes() unless exists  $self->{node_names};

    if ( defined $self->{node_x10_names}->{$x} ) {
	return $self->{node_x10_names}->{$x};
    }

    return undef;
}

sub get_comm_name {
    my $self = shift;
    my $comm = uc( shift );

    #print "get_comm_name : $comm\n";

    $self->load_conf() unless exists  $self->{commands};

    if ( exists $self->{commands}->{$comm}  ) {
	#print "get_comm_name : $comm : $self->{commands}->{$comm}\n";

	return $self->{commands}->{$comm};
    }

    #print "get_comm_name : $comm : $comm\n";
    #print "=== get_comm_name/controls ===\n", Dumper($self->{commands}), "\n";
    return $comm;
}

sub get_node_name {
    my $self = shift;
    my $node = shift ;

    # print "get_node_name : $node\n";

    $self->load_nodes() unless exists  $self->{node_names};

    if ( exists $self->{nodes}->{$node}
	&& exists $self->{nodes}->{$node}->{name} ) {

	return $self->{nodes}->{$node}->{name};
    }

    if ( exists $self->{scenes}->{$node}
	&& exists $self->{scenes}->{$node}->{name} ) {

	return $self->{scenes}->{$node}->{name};
    }

    return $node;
}

sub get_node_id {
    my $self = shift;
    my $n = shift ;
    my $node = uc( $n );

#print "get_node_id : $node\n";

    $self->load_nodes() unless exists  $self->{node_names};

    if ( exists ( $self->{node_names}->{$node} ) ) {
	return ( $self->{node_names}->{$node} );
    }

    if ( exists ( $self->{nodes}->{$node} ) ) {
	return ( $node );
    }

    if ( exists ( $self->{scene_names}->{$node} ) ) {
	return ( $self->{scene_names}->{$node} );
    }

    if ( exists ( $self->{scenes}->{$node} ) ) {
	return ( $node );
    }

    if ( exists ( $self->{scenes}->{$n} ) ) { ### FIX
	return ( $n );
    }

    if ( exists ( $self->{node_x10}->{$node} ) ) {
	return ( $self->{node_x10}->{$node} );
    }


    return undef;
}

sub get_comm_id {
    my $self = shift;
    my $comm = uc( shift );

#print "get_comm_id : $comm\n";

    $self->load_conf() unless exists $self->{commands};

    if ( exists ( $self->{commands}->{$comm} ) ) {
	return ( $comm );
    }

    if ( exists ( $self->{controls}->{$comm} ) ) {
	return ( $self->{controls}->{$comm} );
    }

    return undef;
};

sub error {
my $self = shift;
    return $self->{error};
}

sub clear_error {
my $self = shift;
    $self->{error} = undef; 
}


1;

__END__


get_comm_id : internal
get_node_id : internal
get_prog_id : internal
get_wol_id : internal

load_clim : internal
load_conf : internal
load_netw : internal
load_nodes : internal
load_prog : internal
load_vars : internal

log_query : retrieve log
log_parse : parse log line
log_reset : reset the logs
node_comm : Send a command to an Insteon device ( by name or addr )
node_comm_simple  : Send a command to an Insteon device
node_names : returns a list of Insteon device names / addresses
node_query : return all device properties
node_status : get status of an Insteon device ( node->property->value )
prog_comm : Send a command
prog_names : returns a list of programs
prog_query : return all device properties
prog_status : get status of program by name
var_status : returns a hash of variables and their value.
wol_comm : send an WOL to a registared network host
x10_comm : send an X10 command by X10 addr
