#!/usr/bin/perl

# get and print Climate inro from an ISY99
#
# Peter Shipley <peter.shipley@gmail.com>


use strict;
use warnings;

use ISY::Simple;
use Getopt::Long;
use Data::Dumper;
$Data::Dumper::Indent = 1;         # pretty print with array indices

my ($debug, $logfile,  $raw);

my $isylogin="admin";
my $isypass="admin";
my $ip="10.1.1.36"; #debug
my $verbose=2;
my $printdate;
my @key;
my $newl;

my ($user, $pass);

sub main {
my $help;

    GetOptions ( # 'debug!'	=> \$debug,
		'verbose!'	=> \$verbose,
		'raw'		=> \$raw,
		'ip'            => \$ip, 
		'user'          => \$user,
		'pass'          => \$pass,
		'n!'		=> \$newl,
		'date!'		=> \$printdate,
		'key=s'		=> \@key,
		'help|?:s'	=> \$help,
		);

    if ( defined $help ) {
	usage($help);
	exit;
    }

    @key = split(/,/,join(',',@key));

    $ip = $ENV{"ISY_ADDR"} if (!$ip && exists $ENV{"ISY_ADDR"});
    die "ISY ip address not specified" unless ($ip);

    my $isy = ISY::Simple->new(
	host => $ip,
	user => $user,
	pass => $pass,
	verbose => $verbose,
	faststart => "1",
    );

    unless ( $isy ) {
	print "new failed\n";
	exit(0);
    }

    my $cli = $isy->clim_query();

    unless ( $cli ) {
	print STDERR "Error: ", $isy->{error}, "\n";
	exit;
    }

#    if ( @key ) {
#	for $k ( @key ) {
#	id ( ! exists $cli->{$key} ) {
#	    print "Unknown key : $key\n";
#	}
#    }


    if ( $printdate ) {
	print scalar (localtime(time)), ( @key ? "\t" : "\n" );
    }


    if ( @key ) {
	print join("\t", map { $cli->{$_} }  @key), "\n";
	#for my $k ( @key ) {
	#    print "\t$cli->{$k}" if ( exists $cli->{$k} ) ;
	#}
	#print "\n";
    } else {
	if ( $raw ) {
	    print Dumper($cli), "\n";
	} else {
	    for my $k ( sort keys ( %{$cli} ) ) {
		next if $k eq "rss" ;
		print "$k : $cli->{$k}\n";
	    }
	}
    }

    exit;
}


sub usage {
    my ($h) = @_;
    print "$0 : [-k key1,key2] | -r n",
	"\t-k <key>:\tprint only key\n",
	"\t-r :\traw dump\n",
	"\t-d :\tprint date\n",
	"\n",
	"\t-i :\tIP addr of ISY device\n",
	"\t-u :\tlogin of ISY device\n",
	"\t-p :\tpass of ISY device\n",
	"\n",
	"\t-help :\tprint options\n",
	"\n",
	"\n";
}

&main();
exit(0);


__END__
