#!/usr/bin/perl

# List devices assocated with an ISY unit
#
# Peter Shipley <peter.shipley@gmail.com>

use strict;
use warnings;


use ISY::Simple;
use Getopt::Long;
use Data::Dumper;
$Data::Dumper::Indent = 1;         # pretty print with array indices


my $isylogin="admin";
my $isypass="admin";
my $ip="10.1.1.36"; #debug
my $verbose=0;
my $longfmt=0;

my $listname = 1;
my $listaddr = 1;
my $liststate = 1;
my $listelk = 0;
my $listtype = 0;
my $listgroups = 0;
my $listscenes = 0;
my $listmembers = 0;

my ($user, $pass);

sub main {
my $help;


    GetOptions (
		'ip'            => \$ip, 
		'user'          => \$user,
		'pass'          => \$pass,
		'help|?:s'	=> \$help,
		'verbose+'	=> \$verbose,
		'l+'		=> \$longfmt,
		'name!'		=> \$listname,
		'addr!'		=> \$listaddr,
		'state!'	=> \$liststate,
		'elk!'		=> \$listelk,
		'type!'		=> \$listtype,
		'groups!'	=> \$listgroups,
		'S'		=> \$listscenes,
		'members'	=> \$listmembers,
		);

    # print "ARGV = ", join(' : ', @ARGV), "\n";

    if ( defined $help ) {
	usage($help);
	exit;
    }

    #print " $listname : $listaddr : $liststate : $listtype : $listgroups\n";

    if ( $longfmt ) {
	$listname = $listaddr = $liststate = $listtype = $listgroups = 1;
    }

    if ( $listscenes ) {
	# $listgroups = 1;
	$liststate = 0;
	$listtype = 0;
    }

    #print " $listname : $listaddr : $liststate : $listtype : $listgroups\n";

    $ip = $ENV{"ISY_ADDR"} if (!$ip && exists $ENV{"ISY_ADDR"});
    die "ISY ip address not specified" unless ($ip);

    my $isy = ISY::Simple->new(
	host => $ip,
	user => $user,
	pass => $pass,
	faststart => "1",
    );

    unless ( $isy ) {
	print "new failed\n";
	exit(0);
    }

    my @naddrs;
    if ( @ARGV ) {
	for my $t ( @ARGV ) { 
	    my $naddr = $isy->get_node_id($t);

	    if ( $naddr ) {
		push(@naddrs, $naddr);
		if ( $listmembers ) {
		    push(@naddrs, $isy->scene_nodes($naddr));
		}
	    } else {
		print "unknown node/scene $t\n";
	    }
	}
    } else {
	if ( $listscenes ) {
	    $listgroups = 1;
	    @naddrs = sort $isy->scene_addrs();
	} else {
	    @naddrs = sort $isy->node_addrs();
	}
    }

    my $format = "";
    my @colnames  ;

    if ( $listname ) {
	$format .= "%-20s ";
	push(@colnames, "Name");
    }
    if ( $listaddr ) {
	$format .= "%-12s ";
	push(@colnames, "Addr");
    }

    if ( $liststate ) {
	$format .= "%-4s ";
	push(@colnames, "State");
    }

    if ( $listelk ) {
	$format .= "%-4s ";
	push(@colnames, "ELK");
    }

    if ( $listtype ) {
	$format .= " %-33s";
	push(@colnames, "Type");
    }

    if ( $listgroups ) {
	#$format .= " %-33s";
	$format .= " %s";
	if ( $listscenes ) {
	    push(@colnames, "Members");
	} else {
	    push(@colnames, "Scenes");
	}
    }
    
    $format  .= "\n";

    printf $format , @colnames;

    # print "naddrs = ", join(' : ', @naddrs), "\n";

    for my $a ( @naddrs ) {
	my $n = $isy->get_node($a);
	my @col = ();

	unless ( $n ) { # should not happen 
	    print STDERR "Error: ", $isy->{error}, "\n";
	}

	if ( $listname ) {
	    push(@col, $n->{name});
	}

	if ( $listaddr ) {
	    my $addr;
	    $addr = $isy->x10_node($a) if ($a  =~ /^FF/);
	    $addr = $a unless $addr;
	    push(@col, $addr);
	}

	if ( $liststate ) {
	    if ( exists $n->{property} ) {
		my $state = $n->{property}->{ST}->{formatted};
		$state .= "%" unless ( $state =~ /^[O\s]/ );
		push(@col, $state);
	    } else {
		push(@col, "-");
	    }
	}

	if ( $listelk ) {
	    if ( exists $n->{ELK_ID} ) {
		push(@col, $n->{ELK_ID});
	    } else {
		push(@col, "-");
	    }
	}

	if ( $listtype ) {
	    if ( exists $n->{type} ) {
		if ( $verbose || $longfmt > 1 ) {
		    ## push(@col, join(',', $isy->node_type($n->{type}) ));
		    push(@col, join(',', $isy->node_type($n->{type}), "[$n->{type}]" ));
		} else {
		    #push(@col,  ($isy->node_type($n->{type}))[1] );
		    my @l = $isy->node_type($n->{type});
		    if ( $l[1] ne "UNKNOWN" ) {
			push(@col,  $l[1]);
		    } else {
			push(@col,  $l[0]);
		    }
		}
	    } elsif ( $n->{flag} & 0x04 ) { #  NODE_IS_A_GROUP
		push(@col, "SCENE/GROUP");
	    } else {
		push(@col, "-"); # should not happen
	    }
	}

	if ( $listgroups ) {
	    if ( $n->{flag} & 0x04 ) {
		my @s = $isy->scene_nodes($a);
		push(@col, join(', ', sort map { $isy->get_node_name($_) } @s));
	    } else {
		my @s = $isy->node_scenes($a);
		push(@col, join(', ', sort map { $isy->get_node_name($_) } @s));
	    }
	}


	printf $format, @col;

	# print "=== $a ===\n", Dumper($n), "\n";
    }



    exit;
}


sub usage {
    use Pod::Usage;

#	pod2usage(2);
	pod2usage(-exitstatus => 0, -verbose => 2);
}

&main();
exit(0);


__END__


=head1 NAME

isy_ls

=head1 SYNOPSIS

    isy_ls [--ip] [-l] [file ...]

=over 8

=head1 OPTIONS

=item B<--ip>

IP of ISY

=item B<--user>

login

=item B<--pass>

password

=item B<--verbose>

more verbose output formating

=item B<-l>

long List in long format

=item B<--name|-noname>

print node/scene names

=item B<--addr>

print node/scene address

=item B<--state>

print node state

=item B<--elk>

print elk address

=item B<--type>

print node type

=item B<--groups>

list groups node is member of

=item B<--S>

list scenes

=item B<--members>

list members scenes

=item B<-help>

Print a brief help message and exits.

=back

=head1 DESCRIPTION

B<isy_ls> will query a ISY contoller device and report node and scene infomation

=cut
