#!/usr/bin/perl

# Print log from an ISY unit
#
# Peter Shipley <peter.shipley@gmail.com>

use strict;

use ISY::Simple;
use Getopt::Long;
use Data::Dumper;
$Data::Dumper::Indent = 1;         # pretty print with array indices

my ($debug, $logfile,  $clearlog, $raw);

my $isylogin="admin";
my $isypass="admin";
my $ip="10.1.1.36"; #debug
my $verbose=2;
my $errorlog=0;
my $printlog=0;
my $namenodes=1;
my $lines;

my ($nodefilter, $controlfilter,$actionfilter, $msgfilter);
my @typefilter;
my @userfilter;
my ($user, $pass);

# 2208945600
# 2208988800
my $timeconst=2208988800;

sub main {
my $help;

    GetOptions ('debug:+'	=> \$debug,
		'verbose!'	=> \$verbose,
		'outfile=s'	=> \$logfile,
		'lines=i'	=> \$lines,
		'error!'	=> \$errorlog,
		'ip'		=> \$ip,
		'user'		=> \$user,
		'pass'		=> \$pass,
		'raw'		=> \$raw,
		'namenodes!'	=> \$namenodes,
		'printlog!'	=> \$printlog,  # print before reseting logs
		'clear!'	=> \$clearlog,
		'N|node=s'	=> \$nodefilter,
		'C|control=s'	=> \$controlfilter,
		'A|action=s'	=> \$actionfilter,
		'T|type=s'	=> \@typefilter,
		'U|user=s'	=> \@userfilter,
		'M|message=s'	=> \$msgfilter,
		'help|?:s'	=> \$help,
		);

# log : Node Control Action Time UID Type
# Err :  Time UID Type message

    $printlog=1 unless ( $clearlog );

    if ( defined $help ) {
	usage($help);
	exit;
    }

    $ip = $ENV{"ISY_ADDR"} if (!$ip && exists $ENV{"ISY_ADDR"});
    die "ISY ip address not specified" unless ($ip);

    if ( $logfile ) {
	 open(STDOUT, '>', $logfile) or die "Can't open $logfile: $!";
    }


    @userfilter = split(/,/,join(',',@userfilter));
    @typefilter = split(/,/,join(',',@typefilter));

    #print "userfilter=", join(',',@userfilter), "\n";
    #print "typefilter=", join(',',@typefilter), "\n";

    my $isy = ISY::Simple->new(
	host => $ip,
	user => $user,
	pass => $pass,
	debug => $debug,
	faststart => "1",
    );

    unless ( $isy ) {
	print "new failed\n";
	exit(0);
    }

    if (  $printlog ) {
	print_log($isy, $errorlog);
    }

    if (  $clearlog ) {
	clear_log($isy, $errorlog);
    }

}

sub clear_log  {
    my $isy = shift;
    my $err = shift;

   return  $isy->log_query( error => $err, reset => 1 );;
}



sub print_log  {
    my $isy = shift;
    my $err = shift;

    my $loglines = $isy->log_query( error => $err );;

    my $first_l=0;
    if ( $lines && ( $lines <  $#{$loglines}) ) {
	$first_l = $#{$loglines} - $lines;
    }

    if ($raw) {
	for my $i ( $first_l ... $#{$loglines} ) {
	    print "${$loglines}[$i]\n";
	}
    } else {
	my $tzo = timezone_offset(); 
	my $names;
#	print "tzo = $tzo\n";
	my %ulist;
	my %tlist;
	my $fmt="";

	if ( @userfilter ) {
	    for my $u ( @userfilter ) {
		$ulist{$u}++;
	    }
	}

	if ( @typefilter ) {
	    for my $t ( @typefilter ) {
		$tlist{$t}++;
	    }
	}


	#print "namenodes = $namenodes\n";

	if ( $verbose ) {
	    print "$lines out of " if $lines;
	    print scalar(@{$loglines}), "\n";
	}


	$fmt = "%-5d " if ( $debug ) ;
	my @abbr = qw( Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec );
	my $nodefmt="%-12s";
	my $commfmt="%-4s";

	if ( $err ) { 
		my @label = ( "Time", "UID", "Log Type", "Error Message");
	        $fmt .= "%15s %-24s %-38s : \"%s\"\n";
		unshift(@label, "#") if $debug; 
		printf STDOUT "%-15s  %-24s %-38s :  %s\n",  @label;
	} else {
	    if ( $namenodes ) {
		$nodefmt="%-18s";
		$commfmt="%-10s";
	    }
	    my @label = ( "Node", "Control", "Action", "Time", "UID", "Log Type");
	    unshift(@label, "#") if $debug; 
	    $fmt .= "$nodefmt $commfmt %-20s %-15s  %-24s %s\n";
	    printf STDOUT $fmt, @label;
	#	"Node", "Control", "Action", "Time", "UID", "Log Type";
	    # labelfmt: "$nodefmt $commfmt %-20s %-15s  %-24s %s\n",

	}


	if ( $debug ) {
	    print "nodefilter = \"$nodefilter\" \n" if ( $nodefilter ) ;
	    print "actionfilter = \"$actionfilter\" \n" if ( $actionfilter ) ;
	    #print "tlist = ", join(',', keys %tlist), "\n" if ( @typefilter ) ;
	}

	for my $i ( $first_l ... $#{$loglines})
	{
	    # print "$i:  ${$loglines}[$i]\n" if $debug;
	    my @l = split("\t", ${$loglines}[$i]);
	    my @a;

	    if ( $err ) { # Time UID Type message

		next if ( @userfilter && !exists $ulist{$l[1]});
		next if ( @typefilter && !exists $tlist{$l[2]});


		@a = localtime($l[0] - $timeconst - $tzo);
		$l[0] = sprintf "%s %.2d, %.2d:%.2d:%.2d", $abbr[$a[4]], $a[3], $a[2], $a[1], $a[0];
		#"$abbr[$a[4]] $a[3] $a[2]:$a[1]:$a[0]";
		# $l[0] = scalar( localtime($l[0] - 2208988800 - $tzo));
		my $u =  getuser($l[2]);

		#print "l(1) =  $l[1] , u = \"$u\"\n";

		$l[1] = getuser($l[1]);
		$l[2] = $isy->log_type($l[2]) || $l[2];

		# old "%s %.2d, %.2d:%.2d:%.2d  %-24s %-38s : \"%s\"\n",
		#	$abbr[$a[4]], $a[3], $a[2], $a[1], $a[0], @l; 

		unshift( @l, $i ) if $debug;
		printf STDOUT $fmt, @l;
	    } else {  # Node Control Action Time UID Type

		next if ( @userfilter && !exists $ulist{$l[4]});
		next if ( @typefilter && !exists $tlist{$l[5]});
		next if ( $actionfilter && $l[2] !~ /$actionfilter/ );

		my $nodeaddr = $l[0];
		if ( $namenodes ) {
		    if ( $l[0] =~ /^X10/  ) {
			if ( my $n = $isy->get_x10_name($l[1]) ) {
			    $l[0] .= " ($n)";
			}
		    } else {
			$l[0] = $isy->get_node_name($l[0]);
			$l[1] = $isy->get_comm_name($l[1]);
		    }
		}

		next if ( $nodefilter && ( $l[0] !~ /$nodefilter/io && $nodeaddr !~ /$nodefilter/io ) );


		# $l[3] = scalar( localtime($l[3] - 2208988800 - $tzo));
		my @a = localtime($l[3] - $timeconst - $tzo);
		$l[3] = sprintf "%s %.2d, %.2d:%.2d:%.2d", $abbr[$a[4]], $a[3], $a[2], $a[1], $a[0];
		$l[4] = getuser($l[4]);
		$l[5] = $isy->log_type($l[5]) || $l[5];

		unshift( @l, $i ) if $debug;
		printf STDOUT $fmt, @l; 
		#printf STDOUT "$nodefmt $commfmt %-20s %s  %-24s %s\n", @l; 
		#"$i :  ", join("\t", @l), "\n";
	    }
	    
	}
    }
}


my @userid = ( "SYSTEM_USER", "SYSTEM_DRIVER_USER", "WEB_USER", "SCHEDULER_USER", "D2D_USER", "ELK_USER",
        "SEP_DEVICE_UMETER_USER", "SEP_DEVICE_UPRICE_USER", "SEP_DEVICE_UMSG_USER", "SEP_DEVICE_UDR_USER", "GAS_METER_USER");


		  
sub getuser {
    my ($u) = @_;
     
     if ( $u >= 0 && $u <= $#userid ) {
	 #print "getuser : $u =  $userid[$u]\n";
	 return $userid[$u];
     }

     return $u;
}






sub timezone_offset {
    # return -28800;

    my @lctime = localtime();
    my @gmtime = gmtime();
    # index 2 represent the hours
    # index 8 represent isdst (daylight saving time boolean (0/1))

     return (($lctime[2] ) - ($gmtime[2] - $gmtime[8])) * 60 * 60;
    # return (($lctime[2] - $lctime[8]) - ($gmtime[2] - $gmtime[8])) * 60 * 60;
}




sub usage {
    my ($h) = @_;
    print "$0 : [-l #] [-c] [-e] [-r] [-n] [-u loguid] [-t logtype] [-i <addr>]\n",
	"\t-p :\tprints log ( default )\n",
	"\t-c :\tclears log\n",
	"\t-e :\tprint error log\n",
	"\t-r :\tprint raw log\n",
	"\t-n :\tprint node names instead of addresses ( default )\n",
	"\t-l <n>:\tlist last <n> lines\n",
	"\n",
	"\t--user :\tlogin of ISY device\n",
	"\t--pass :\tpass of ISY device\n",
	"\t-i :\tIP addr of ISY device \n",
	    "\t\t( default taken from env var ISY_ADDR\n",
	"\n",
	"\t-N|Node <s>:\tlist of lines refering to node <regex>\n",
	"\t-C|control <s>:\tlist of lines refering a control <regex>\n",
	"\t-A|action <s>:\tlist of lines refering a action <regex>\n",
	"\t-T|type <s>:\tlist of lines refering a log type <string>\n",
	    "\t\tSee -help types for a list of valid types\n",
	"\t-U|user <s>:\tlist of lines refering a log user uid <string>\n",
	    "\t\tSee -help users for a list of valid uids\n",
	"\t-M|message <s>:\tfilter lines my message string <regex> ( error log only) \n",
	"\n",
	"\t-help :\tprint options\n",
	"\t-help users :\tprint loguser list\n",
	"\t-help types :\tprint logtype list\n",
	"\n";


# log : Node Control Action Time UID Type
# Err :  Time UID Type message

    $printlog=1 unless ( $printlog || $clearlog );

    if ($h eq "users") {
	print "userid =";
	for my $i ( 0 ... $#userid )  {
	    print " $i=$userid[$i]";
	}
	print "\n";
	return;
    }

    my @l = ( undef, "SYSTEM_STARTUP", "SYSTEM_SHUTDOWN", "WARNING", "INFO", "LOG" );
    if ($h eq "types") {
	print "types =";
	 for my $i ( 1 ... 5 ) {
	    print " $i=$l[$i]"
	}
	print "\n";
	return;
    }

}

&main();
exit(0);


__END__
