#!/usr/bin/perl

# List or set Program Variables on an ISY
#
# Peter Shipley <peter.shipley@gmail.com>

use strict;

use ISY::Simple;
use Getopt::Long;
# use Data::Dumper;
# $Data::Dumper::Indent = 1;         # pretty print with array indices

my ($logfile,  $clearlog, $raw);

my $debug=0;
my $isylogin="admin";
my $isypass="admin";
my $ip="10.1.1.36"; #debug
my $verbose=2;
my $list;
my $init;
my $quiet=0;
my $name=0;

my ($user, $pass);

sub main {
my $help;

    GetOptions ('debug+'	=> \$debug,
		'verbose!'	=> \$verbose,
		'ip'            => \$ip, 
		'quiet'		=> \$quiet,
		'user'          => \$user,
		'pass'          => \$pass,
		'list'		=> \$list,
		'I|init'	=> \$init,
		'name'		=> \$name,
		'help|?:s'	=> \$help,
		);

    if ( defined $help ) {
	usage($help);
	exit;
    }

    $ip = $ENV{"ISY_ADDR"} if (!$ip && exists $ENV{"ISY_ADDR"});
    die "ISY ip address not specified" unless ($ip);

    my $isy = ISY::Simple->new(
	host => $ip,
	user => $user,
	pass => $pass,
	verbose => $verbose,
	faststart => "1",
	debug => $debug,
    );

    unless ( $isy ) {
	print "new failed\n";
	exit(0);
    }

    unless ( $isy->load_vars() ) {
	print STDERR "Error: ", $isy->{error}, "\n";
	exit(0);
    }
    

    if (  $list or ! @ARGV ) {
	list_vars($isy);
	exit(0);
    }

    # $isy->load_vars();


    for my $targ ( @ARGV ) {
    my $val;

	if ( $targ =~ /([^=]+)=(\d+)/ ) {
	    #print "var_set $1 = $2\n";
	    $val = $isy->var_set($1, $2, undef) ;
	    print STDERR "Error: ", $isy->{error}, "\n" unless $val;

	} else {

	    $val = $isy->var_get($targ) ;

	    print STDERR "Error: ", $isy->{error}, "\n"  unless defined $val;

	    if ( $name ) {
		print "$targ=$val\n";
	    } else {
		print $val, "\n";
	    }

	}
    }
}

sub list_vars  {
    my $isy = shift;

    my $vars = $isy->var_query($isy);

    return unless $vars;

    my $pfmt = "%-4s : %-19s%-5s\t%-5s\t%s\n";

    printf $pfmt, "ID", "NAME", "VAL", "INIT", "DATE";
    for my $t ( 1 , 2) {
	for my $k ( sort { $a <=> $b } keys ( %{$vars->{$t}} ) ) {
	    printf $pfmt, "$t:$k",
		$vars->{$t}->{$k}->{name},
		$vars->{$t}->{$k}->{val},
		$vars->{$t}->{$k}->{init},
		$vars->{$t}->{$k}->{ts};
	}
    }

   return  ;
}


sub usage {
    my ($h) = @_;
    print "$0 : [-l] | host\n",
	"\t-l <n>:\tlist registarted WOL ids and names\n",
	"\n",
	"\t-i :\tIP addr of ISY device\n",
	"\t-u :\tlogin of ISY device\n",
	"\t-p :\tpass of ISY device\n",
	"\t-l :\tlist vars\n",
	"\t-I :\tinit\n",
	"\t-n :\tname\n",
	"\n",
	"\t-help :\tprint options\n",
	"\n",
	"\n";
}

&main();
exit(0);


__END__
