#!/usr/bin/perl

# List or Sed a WoL (Wake on Lan) command from an ISY
#
# Peter Shipley <peter.shipley@gmail.com>

use strict;

use ISY::Simple;
use Getopt::Long;
# use Data::Dumper;
# $Data::Dumper::Indent = 1;         # pretty print with array indices

my ($debug, $logfile,  $clearlog, $raw);

my $isylogin="admin";
my $isypass="admin";
my $ip="10.1.1.36"; #debug
my $verbose=2;
my $list;
my $quiet=0;

my ($user, $pass);

sub main {
my $help;

    GetOptions ('debug!'	=> \$debug,
		'verbose!'	=> \$verbose,
		'ip'            => \$ip, 
		'quiet'		=> \$quiet,
		'user'          => \$user,
		'pass'          => \$pass,
		'list'		=> \$list,
		'help|?:s'	=> \$help,
		);

    if ( defined $help ) {
	usage($help);
	exit;
    }

    $ip = $ENV{"ISY_ADDR"} if (!$ip && exists $ENV{"ISY_ADDR"});
    die "ISY ip address not specified" unless ($ip);

    my $isy = ISY::Simple->new(
	host => $ip,
	user => $user,
	pass => $pass,
	verbose => $verbose,
	faststart => "1",
    );

    unless ( $isy ) {
	print "new failed\n";
	exit(0);
    }

    unless ( $isy->load_netw() ) {
	print STDERR "Error: ", $isy->{error}, "\n";
	exit(0);
    }
    

    if (  $list ) {
	list_wol($isy);
	exit(0);
    }

    for my $targ ( @ARGV ) {

	my $wid;
	if ( $wid =  $isy->wol_comm( $targ) ) {
	    print "WOL sent to $wid->{name} ( $wid->{mac} )\n" unless $quiet;
	} else {
	    print STDERR "Error: ", $isy->{error}, "\n";
	    next;
	}
    }


}

sub list_wol  {
    my $isy = shift;

    my $wol = $isy->wol_query($isy);

    return unless $wol;

    printf "%s : %-19s\t%s\n", "ID", "MAC", "NAME";
    for my $k ( sort keys ( %{$wol} ) ) {
	printf "%-2d : %-19s\t%s\n", $k,
	    $wol->{$k}->{mac},
	    $wol->{$k}->{name} ;
    }

   return  ;
}


sub usage {
    my ($h) = @_;
    print "$0 : [-l] | host\n",
	"\t-l <n>:\tlist registarted WOL ids and names\n",
	"\n",
	"\t-i :\tIP addr of ISY device\n",
	"\t-u :\tlogin of ISY device\n",
	"\t-p :\tpass of ISY device\n",
	"\n",
	"\t-help :\tprint options\n",
	"\n",
	"\tA host can be a WOLID or registared hostname\n",
	"\n";
}

&main();
exit(0);


__END__
